# zbot/__init__.py
#
#

""" creators outside creation. """

__version__ = 20

## IMPORTS 

from zbot.generic import *
from zbot.utils import *

import collections
import threading
import logging
import _thread
import socket
import queue
import types
import errno
import select
import fcntl
import uuid
import json
import time
import imp
import sys
import os
import re

## ERRORS

class Error(BaseException): pass

class OverloadError(Error): pass

class MissingArgument(Error): pass

class MissingOutFunction(Error): pass

class NoText(Error): pass

class NoDate(Error): pass

class NoPath(Error): pass

class NoFileName(Error): pass

class SignatureError(Error): pass

class NoSuchBotType(Error): pass

class NoEvent(Error): pass

class RemoteDisconnect(Error): pass

class NotOne(Error): pass

class TargetMissing(Error): pass

## BASE

class Object(dict):

    def __getattribute__(self, *args, **kwargs):
        name = args[0]
        if name == "what": return get_cls(self)
        if name == "modname": return self.__class__.__module__
        if name == "cfrom": return called_from(2)
        if name == "stdin": return sys.stdin
        if name == "stdout": return sys.stdout
        if name == "days": return str(datetime.timedelta(-1, time.time() - emailstamp(self.Date)))
        return dict.__getattribute__(self, *args, **kwargs)

    def __getattr__(self, name):
        try: return self[name]
        except KeyError:
            if name == "_result": self["_result"] = Object()
            if name == "_status": self["_status"] = Object()
            if name == "_ready": self["_ready"] = threading.Event()
            if name == "ctime": self["ctime"] = time.time()
        try: return self[name]
        except KeyError: return ""

    def __setattr__(self, name, value): return dict.__setitem__(self, name, value)

    def __exists__(self, a):
        try: return self[a]
        except KeyError: False

    def __lt__(self, a): return self.ctime < a.ctime

    ## locators

    def get_root(self, *args, **kwargs): return j(os.path.expanduser("~"), kernel.config.workdir)

    def get_path(self, *args, **kwargs): return "%s%s%s-=-%s" % (self.get_root(), os.sep, self.what, rtime())

    ## output

    def make_json(self, *args, **kwargs): return json.dumps(self.reduced(), default=smooth, *args, **kwargs)

    def make_full(self, *args, **kwargs): return json.dumps(self, default=smooth, *args, **kwargs)

    def make_signature(self, sig=None): return str(hashlib.sha1(bytes(str(sig or self), "utf-8")).hexdigest())

    def make_path(self, *args, **kwargs): path = self.get_path() ; make_dir(path) ; return path

    ## input

    def load(self, *args, **kwargs):
        if args: path = self.make_path(*args, **kwargs)
        else: path = self.make_path()
        return self.load_file(path)

    def load_file(self, *args, **kwargs):
        path = args[0]
        ondisk = self.read(path) 
        fromdisk = json.loads(ondisk)
        if "data" in fromdisk: self.update(fromdisk["data"])
        return self

    def load_json(self, *args, **kwargs):
        try: self.update(json.loads(args[0]))
        except ValueError: pass
        except: error()
        return self

    def read(self, *args, **kwargs):
        logging.info("read %s" % args[0])
        path = args[0]
        try: f = open(path, "r")
        except IOError as ex:
            if ex.errno == errno.ENOENT: return "{}"
            raise
        if self.do_test: f.line_buffering = False
        res = ""
        for line in f:
            if not line.strip().startswith("#"): res += line
        if not res.strip(): return "{}"
        f.close()
        return res

    ## persistence

    def dump(self, *args, **kwargs):
        if args: path = args[0]
        else: path = self.make_path(*args, **kwargs)
        todisk = Object()
        todisk.data = self.reduced()
        todisk.data.path = path
        todisk.data.day = path.split(os.sep)[-2]
        todisk.create_type = self.what
        todisk.modname = self.modname
        todisk.saved_from = called_from()
        todisk.version = __version__
        todisk.signature = self.make_signature()
        try: result = todisk.make_json(indent=2, ensure_ascii=False)
        except TypeError: raise NoJSON(todisk)
        return result

    def save(self, *args, **kwargs):
        path = self.make_path(*args, **kwargs)
        self.sync(path)
 
    def sync(self, *args, **kwargs):
        if args: path = args[0]
        else: path = self.path
        todisk = self.dump(path, **kwargs)
        datafile = open(path + ".tmp", 'w')
        fcntl.flock(datafile, fcntl.LOCK_EX | fcntl.LOCK_NB)
        datafile.write(headertxt % (path, __version__, "%s characters" % len(todisk)))
        datafile.write(todisk)
        datafile.write("\n")
        fcntl.flock(datafile, fcntl.LOCK_UN)
        datafile.close()
        os.rename(path + ".tmp", path)
        logging.warn("saved %s" % path)
        return self

    ## state

    def register(self, *args, **kwargs):
        name = args[0]
        obj = args[1]
        self[name] = obj

    def reduced(self):
        res = Object()
        for key in self.keys():
            k = str(key)
            if k.startswith("_"): continue   
            if key in ["args", "rest", "first"]: continue
            res[key] = self[key]
        return res

    ## objectors

    def objects(self, *args, **kwargs):
        path = self.get_root(*args, **kwargs)
        res = []
        for p in os.listdir(path):
            fnn = j(path, p)
            if os.path.isdir(fnn): res.extend(self.objects(fnn)) ; continue
            obj = Object().load_file(fnn)
            go = True
            for opt in kwargs:
                try:
                    if kwargs[opt] not in str(obj[opt]): go = False 
                except KeyError: go = False
            if go: res.append(obj)
        return res

    ## result

    def prepare(self, *args, **kwargs):
        try: self.ucmnd, self.rest = self.txt.split(" ", 1)
        except: self.ucmnd = self.txt ; self.rest = ""
        if self.ucmnd and self.ucmnd[0] == ".": self.ucmnd = self.ucmnd[1:]
        self.opts = get_opts(self.rest)
        self.args = get_args(self.rest)

    def ready(self): self._ready.set()

    def wait(self, sec=180.0): self._ready.wait(sec) 

    def reply(self, value):
        if self.outer: o = self.outer ; o.write(value) ; o.write("\n") ; o.flush() ; return
        self._result[time.time()] = value

    def show_items(self): return ["%s=%s" % (a, b) for a, b in self.items() if b]

    def _raw(self, txt): self._target._raw(txt)

    def say(self, channel, txt): self._target.say(channel, txt)

    def display(self, *args, **kwargs):
        for res in sorted(self._result.keys()): self.say(self.channel or "", self._result[res])

    def put(self, *args, **kwargs): self._target.put(*args, **kwargs)

    def handle(self, *args , **kwargs): return resolve(self, *args, **kwargs)

## TASKS

class Runner(threading.Thread):

    def __init__(self, *args, **kwargs):
        threading.Thread.__init__(self, None, self._loop, "thread.%s" % str(time.time()), args, kwargs)
        self.setDaemon(True)
        self._queue = queue.Queue()
        self._state = "idle"

    def _loop(self, *args, **kwargs):
        logging.warn("start %s" % str(self))
        while self._state in ["running", "idle", "callback", "once"]:
            a, kw = self._queue.get()
            if self._state == "stop": break
            try: func = a[0] ; func(*a[1:], **kw)
            except: error()
            time.sleep(0.001)
        logging.warn("stop %s" % str(self))

    def stop(self, *args, **kwargs): self._state = "stop"
        
    def put(self, *args, **kwargs): self._queue.put((args, kwargs)) ; return 

class TaskRunner(Runner):

    def _loop(self, *args, **kwargs):
        logging.warn("start %s" % str(self))
        while self._state in ["running", "idle", "callback", "once"]:
            a, kw = self._queue.get()
            if self._state == "stop": break
            kernel.cb.handle_cb(a[0])
            time.sleep(0.001)
        logging.warn("stop %s" % str(self))

class Spider(Runner):

    def __init__(self, *args, **kwargs):
        self.errors = []
        self.urls = []
        self.followed = []

    def crawl(self, *args, **kwargs):
        event = args[0]
        url = event.url ; depth = event.depth ; speed = event.speed
        if depth < 0: return
        if not self.url.base in url: logging.warn("skipping %s (%s)" % (url, self.url.base)) ; return
        if url in self.errors: logging.warn("skipping %s" % url) ; return
        urls = [] ; linknr = 0 ; follownr = 0 ; n = 0
        if url not in self.urls:
            self.urls.append(url)
            page = Url(url)
            time.sleep(speed)
            logging.warn("fetching %s" % url)
            content = page.fetch()
            logging.warn("status of %s: %s" % (url, content.status))
            try: o = Object() ; o._target = event._target ; o.txt = strip_html(content.read()) ; o.save()
            except Exception as ex: error()
            for p in page.geturls():
                if not p in self.errors:
                    o = Object() ; o.url = p ; o.depth = depth - 1 ; o.speed = speed - 1
                    self.put(self.do_fetch, o)
        if not self._queue.qsize(): self.stop()

    def parse(self, url=None):
        """
           
        Attribute       Index   Value                   Value if not present
        scheme          0       URL scheme specifier    empty string
        netloc          1       Network location part   empty string
        path            2       Hierarchical path       empty string
        query           3       Query component         empty string
        fragment        4       Fragment identifier     empty string
        
        """
        if url: self.url = url
        self.parsed = urllib.parse.urlsplit(url or self.url)
        self.target = self.parsed[2].split("/")
        if "." in self.target[-1]:
            self.basepath = "/".join(self.target[:-1])
            self.file = self.target[-1]
        else: self.basepath = self.parsed[2] ; self.file = None
        if self.basepath.endswith("/"): self.basepath = self.basepath[:-1]
        self.base = urllib.parse.urlunsplit((self.parsed[0], self.parsed[1], self.basepath , "", ""))
        self.root = urllib.parse.urlunsplit((self.parsed[0], self.parsed[1], "", "", ""))

    def geturls(self):
        if not self.response: self.fetch()
        import bs4
        s = bs4.BeautifulSoup(self.txt)
        urls = []
        tags = s('a')
        for tag in tags:
           href = tag.get("href")
           if href:
               href = href.split("#")[0]
               if not href: continue
               if not href.endswith(".html"): continue
               if ".." in href: continue
               if href.startswith("mailto"): continue
               if not "http" in href:
                    if href.startswith("/"): href = self.root + href
                    else: href = self.base + "/" + href
               if not self.root in href: logging.warn("%s not in %s" % (self.root, href)) ; continue
               if href not in urls: urls.append(href)
        logging.warn("found %s urls" % len(urls))
        return urls
        
## DISPATCH

class Dispatcher(Object):

    max = 50
    runners = collections.deque() 

    def stop(self, name=None):
        for runner in self.runners:
            if name and name in runner.name: runner.stop()

    def put(self, *args, **kwargs):
        if not args: raise NoTask()
        target = self.get_target()
        target.put(*args, **kwargs)
        return args[0]

    def get_target(self):
        target = None
        for runner in self.runners:
            if runner._queue and runner._state == "idle": target = runner
        if not target: target = self.makenew()
        return target

    def makenew(self, *args, **kwargs):
        if len(self._runners) < self.max:
            runner = TaskRunner(*args, **kwargs)
            runner.start()
            self.runners.append(runner)
        else: runner = random.choice(self._runners)
        return runner

    def cleanup(self, dojoin=False):
        todo = []
        for runner in self.runners:
            if runner.stopped or not len(runner.queue): todo.append(runner)
        for runner in todo: runner.stop() ; self.runners.remove(runner)

## CALLBACKS

class Callbacks(Dispatcher):

    def register(self, cbtype, cb, *args, **kwargs):
        logging.info("register %s - %s" % (cbtype, str(cb))) 
        if cbtype not in self: self[cbtype] = []
        self[cbtype].append(cb)

    def handle_cb(self, *args, **kwargs):
        event = args[0]
        logging.info(event)
        target = get_cls(event)
        logging.info("cb %s" % target)
        functions = []
        try: functions.extend(self[target])
        except KeyError: pass
        for func in functions:
            try: pre = getattr(func, "pre")
            except AttributeError: pre = None
            if pre and not pre(event): logging.debug("pre %s" % str(func)) ; return
            try: result = func(event)
            except: error()
        return event

## BOTS

class Bot(Callbacks): 

    def connect(self, *args, **kwargs): pass

    def exit(self, *args, **kwargs):
        try: sys.stdout.flush() 
        except Exception as ex: logging.warn(str(ex))
        sys.stdout.close()
        os._exit(0)

    def get_one(self): return None

    def read_some(self, *args, **kwargs): pass

    def put_some(self, *args, **kwargs):
        input = args[0]
        [i, o, e] = select.select([input,], [], [], 0.05)
        gotcha = False
        if i:
            text = i[0].read()
        return gotcha

    def get_prompt(self, *args, **kwargs): return ""

    def begin(self, *args, **kwargs): 
        e = self.get_one(" ".join(self.args))
        self.do_one(e)
        if self.do_shell: self.run_forever(*args, **kwargs)

    def run_forever(self, *args, **kwargs):
        self.connect()
        if not self._state: self._state = "running"
        if self.loop: _thread.start_new_thread(self.loop, ())
        while self._state in ["running", "once"]:
            time.sleep(0.001)
            e = self.get_one()
            if e: self.do_one(e)
            if self._state == "once": break
        self.ready()
 
    def do_one(self, *args, **kwargs):
        ok = True
        try: event = args[0] ; self.put(event) ; event.wait()
        except KeyboardInterrupt: ok = False
        except IndexError: ok = False
        except Exception: error() ; ok = False
        return ok

## PLUGINS

class Plugins(Object):

    def get_names(self, plugsdir): return [x[:-3] for x in os.listdir(plugsdir) if x.endswith(".py")]

    def load_plugs(self, path):
        logging.warn("plugins %s" % path)
        for plugname in self.get_names(path):
            if "__" in plugname: continue
            try: mod = self.load_mod(plugname, path, force=True)
            except: error() ; continue

    def load_package(self, modname):
        mod = __import__(modname)
        path, fn = os.path.split(mod.__file__)
        path += os.sep + "plugs"
        self.load_plugs(path)        

    def load_mod(self, plugname, pdir="", force=False):
        logging.info("load %s - %s" % (self.what.lower(), plugname))
        if plugname in self:
            if not force: return self[plugname]
            self[plugname] = imp.reload(self[plugname])
        else:
            if not pdir: pdir = j(self.root, "plugs")
            search = imp.find_module(plugname, [pdir,])
            self[plugname] = imp.load_module(plugname, *search)
        self.plug_exec(plugname, "init")
        return self[plugname]

    def plug_exec(self, plugname, item):
        try: todo = getattr(self[plugname], item) ; todo() ; logging.info("exec %s" % get_name(todo))
        except AttributeError: pass

    def unload(self, plugname):
        self.plug_exec(plugname, "shutdown")
        del self[plugname]

    def reload(self, plugname, force=False):
        self.unload(plugname)
        mod = self.load_mod(plugname, force)
        return mod

## BOOT

def boot(defaultdir=""):
    global kernel
    config = kernel.config
    try: config.opts, config.args = make_opts()
    except SystemExit: os._exit(1)
    config.update(vars(config.opts))
    if config.do_shell: hello("ZBOT")
    for arg in config.args:
        try: var, val = arg.split("=") ; config[var] = val ; continue
        except ValueError: pass
    if config.do_local: config.workdir = j(os.getcwd(), "workdir") 
    if not config.workdir: config.workdir = defaultdir or "zbot.data"
    make_dir(config.workdir)
    if not config.loglevel: config.loglevel = "error"
    from zbot.log import log_config
    log_config(config.loglevel)
    if config.loglevel:
        logging.warn("E G G S")
        logging.warn("")
        show_eggs()
    if config.loglevel:
        logging.warn("")
        logging.warn("C O N F I G")
        logging.warn("")
        for line in config.show_items(): logging.warn(line)
    if config.loglevel:
        logging.warn("")
        logging.warn("B O O T")
        logging.warn("")
    kernel.cb.register("Object", handle_loop)
    return config

## HANDLER

def handle_loop(event): event.prepare() ; event.handle() ; event.display() ; event.ready()

## VARS

kernel = Object()
kernel.plugins = Plugins()
kernel.cb = Callbacks()
kernel.cmnds = Callbacks()
kernel.tests = Callbacks()
kernel.config = Object()
