# zbot/utils.py
#
#

""" utils package. """

## IMPORT

from zbot import __version__
from zbot.defines import *

from queue import Queue, Empty as QueueEmpty
from traceback import format_exc
from collections import deque 
from cgi import escape

import urllib.request, urllib.error, urllib.parse
import urllib.parse
import traceback
import mailbox
import datetime
import optparse
import _thread
import hashlib
import logging
import urllib
import string
import email
import http
import time
import math
import sys
import os
import re

# AGENT

def useragent(): return 'Mozilla/5.0 (X11; Linux x86_64); ZBOT %s; http://pikacode.com/milla/zbot)' % __version__

## ISTR

class istr(str): pass

## SIGNATURES

def make_signature(data): return str(hashlib.sha1(bytes(str(data), "utf-8")).hexdigest())

## ERRORS

def error(*args, **kwargs):
    msg = exceptionmsg()
    logging.error("error detected:\n\n%s\n" % msg)
    return msg

def exceptionmsg(*args, **kwargs):
    exctype, excvalue, tb = sys.exc_info()
    trace = traceback.extract_tb(tb)
    result = ""
    for i in trace:
        fname = i[0]
        linenr = i[1]
        func = i[2]  
        plugfile = fname[:-3].split(os.sep)
        mod = []
        for i in plugfile[::-1]: mod.append(i)
        ownname = '.'.join(mod[::-1])
        result += "%s:%s %s | " % (ownname, linenr, func)
    del trace
    return "%s%s: %s" % (result, exctype, excvalue)

## JOINS

def j(*args):
     if not args: return
     todo = list(map(str, filter(None, args)))
     return os.path.join(*todo)

def mj(*args):
     if not args: return
     todo = list(map(str, filter(None, args)))
     return os.path.join(*todo).replace(os.sep, ".")

def dj(*args):
     if not args: return
     todo = list(map(str, filter(None, args)))
     return os.path.join(*todo).replace(os.sep, "-")

def aj(sep=None, *args): return os.path.abspath(*j(sep, *args))


## RESOLVING

def resolve_ip(hostname=None, timeout=1.0):
    oldtimeout = socket.getdefaulttimeout()
    socket.setdefaulttimeout(timeout)
    try: ip = socket.gethostbyname(hostname or socket.gethostname())
    except socket.timeout: ip = None
    socket.setdefaulttimeout(oldtimeout)
    return ip

def resolve_host(ip=None, timeout=1.0):
    oldtimeout = socket.getdefaulttimeout()
    socket.setdefaulttimeout(timeout)
    try: host = socket.gethostbyaddr(ip or resolve_ip())[0]
    except socket.timeout: host = None
    socket.setdefaulttimeout(oldtimeout)
    return host

## DIRECTORIES

def touch(fname):
    try: fd = os.open(fname, os.O_RDONLY | os.O_CREAT) ; os.close(fd)
    except: error()

def check_permissions(ddir, dirmask=dirmask, filemask=filemask):
    uid = os.getuid()
    gid = os.getgid()
    try: stat = os.stat(ddir)
    except OSError: make_dir(ddir) ; stat = os.stat(ddir) 
    if stat.st_uid != uid: os.chown(ddir, uid, gid)
    if os.path.isfile(ddir): mask = filemask
    else: mask = dirmask
    if stat.st_mode != mask: os.chmod(ddir, mask)

def make_dir(path):
    target = os.sep
    for item in path.split(target)[:-1]:
        target = j(target, item)
        try: os.mkdir(target)
        except OSError as ex: logging.debug(ex) ; continue
        check_permissions(target)
    return path

## STACK

def dump_frame(search="code"):
    result = {}
    frame = sys._getframe(1)
    search = str(search)
    for i in dir(frame):
        if search in i:
            target = getattr(frame, i)
            for j in dir(target):
                result[j] = getattr(target, j)
    return result

def called_from(level=2):
    result = ""  
    loopframe = sys._getframe(level)
    if not loopframe: return result
    marker = ""
    pre = "zbot"
    while 1:
        try: back = loopframe.f_back
        except AttributeError: break
        if not back: break
        filename = back.f_code.co_filename
        if "plugs" in filename: result = filename.split(os.sep)[-1][:-3] ; pre = "plugs" ; break
        loopframe = back
    del loopframe   
    if result: return "%s.%s" % (pre, result)

## LOCATING

def get_source(mod):
    if not os.getcwd() in sys.path: sys.path.insert(0, os.getcwd())
    source = None
    splitted = mod.split(".")
    if len(splitted) == 1: splitted.append("")
    thedir, file = os.path.split(mod.replace(".", os.sep))
    if os.path.isdir(thedir): source = thedir
    if source and os.path.exists(source): logging.info("source is %s" % source) ; return source
    if not source:
        try: import pkg_resources
        except (ImportError, ValueError): import wtf.contrib.pkg_resources
        source = p.resource_filename()
    logging.info("source is %s" % source)
    return source

## HELPERS

def stripbadchar(s): return "".join([c for c in s if ord(c) > 31 or c in allowedchars])

def enc_char(s):
    result = []
    for c in s:
        if c in allowedchars: result.append(c)
        else: result.append(enc_name(c))
    return "".join(result)

def enc_needed(s): return [c for c in s if c not in allowedchars]

def enc_name(input): return str(base64.urlsafe_b64encode(bytes(input, "utf-8")), "utf-8")

def split_txt(what, l=375):
    txtlist = []
    start = 0
    end = l
    length = len(what)
    for i in range(int(length/end+1)):
        starttag = what.find("</", end)
        if starttag != -1: endword = what.find('>', end) + 1
        else:
            endword = what.find(' ', end)
            if endword == -1: endword = length
        res = what[start:endword]
        if res: txtlist.append(res)
        start = endword
        end = start + l
    return txtlist

def smooth(a):
    if type(a) not in basic_types: return get_cls(a)
    else: return a

def make_version(name=""): return "%s%s #%s -=- ! %s%s" % (YELLOW, name, __version__, time.ctime(time.time()), ENDC)

def hello(name=""): print(make_version(name) + "\n")

def get_opts(input):
    from zbot import Object
    opts = Object()
    for opt in input.split():
        try: name, value = opt.split("=") ; opts[name] = value
        except: pass
    return opts

def get_args(input):
    args = []
    for opt in input.split():
        try: name, value = opt.split("=") ; args.append(name)
        except ValueError: args.append(opt)
        except: error()
    return args

def list_eggs(filter=""):
    for f in sys.path:
        if ".egg" not in f: continue
        if filter and filter not in f: continue
        yield f

def show_eggs(filter="zbot"):
    for egg in list_eggs(filter): logging.warn("%s egg: %s" % (filter, egg))

def stripped(input):
    try: return input.split("/")[0]
    except: return input

def get_clsname(obj):
    name = str(obj.__class__)
    return name.split(" ")[1][1:-2]

def get_cls(obj):
    return get_clsname(obj).split(".")[-1]

## TIME

def dtime(stamp): return datetime.datetime.fromtimestamp(stamp)
def ptime(daystr): return datetime.datetime.strptime(daystr or "1984-01-01", '%Y-%m-%d')
def tdiff(d1, d2): return datetime.timedelta(d1, d2)
def rtime(): return str(datetime.datetime.now()).replace(" ", "-=-")

def hms(): return str(datetime.datetime.today()).split()[1].split(".")[0]
def day(): return str(datetime.datetime.today()).split()[0]

def emailstamp(input):
    from zbot import NoDate
    instring = " ".join(input.strip().split()[:-1])
    try: ttime = time.strptime(instring, "%a, %d %b %Y %H:%M:%S")
    except ValueError: return NoDate()
    return datetime.datetime.fromtimestamp(time.mktime(ttime))

## HEADER

headertxt = '''# %s
#
# this is an zbot (#%s) file, %s
#
# this file can be edited !!

'''

## FEEDER

def feed(text):
    from zbot import Object
    result = []
    chunks = text.split("\r\n")
    for chunk in chunks:
        obj = Object().feed(chunk)
        result.append(obj)
    return result

## PARSER

def parse_email(fn):
    from zbot import Object
    f = open(fn ,"r", errors="replace", encoding="utf-8")
    mails = []
    result = []
    mess = ""
    nr = 0
    for line in f:
        if line.startswith("From "): mails.append(mess) ; mess = line ; continue
        mess += line
    for mess in mails:
        m = email.message_from_string(mess)
        o = Object()
        o.update(m.items())
        o.text = ""
        for load in m.get_payload(): o.text += str(load)
        result.append(o)
    logging.warn("%s emails read" % len(result))
    return result[1:]


## STRIPHTML

def strip_html(text):
    from bs4 import BeautifulSoup
    soup = BeautifulSoup(text)
    result = soup.findAll()
    return str(result)

## STRIPBRACKET

def strip_dingest(text):
    text = urllib.parse.unquote_plus(text)
    text = text.replace("[", "")
    text = text.replace("[[", "")
    text = text.replace("]", "")
    text = text.replace("]]", "")
    text = text.replace("&gt;", ">")
    text = text.replace("&lt;", "<")
    text = text.replace("}}", "")
    text = text.replace("{{", "")
    text = text.replace("{", "")
    text = text.replace("}", "")
    return text

## ENCODING

def get_encoding(data):
    if hasattr(data, 'info') and 'content-type' in data.info and 'charset' in data.info['content-type'].lower():
        charset = data.info['content-type'].lower().split('charset', 1)[1].strip()
        if charset[0] == '=':
            charset = charset[1:].strip()
            if ';' in charset: return charset.split(';')[0].strip()
            return charset
    if '<meta' in data.lower():
        metas = re.findall('<meta[^>]+>', data, re.I | re.M)
        if metas:
            for meta in metas:
                test_http_equiv = re.search('http-equiv\s*=\s*[\'"]([^\'"]+)[\'"]', meta, re.I)
                if test_http_equiv and test_http_equiv.group(1).lower() == 'content-type':
                    test_content = re.search('content\s*=\s*[\'"]([^\'"]+)[\'"]', meta, re.I)
                    if test_content:
                        test_charset = re.search('charset\s*=\s*([^\s\'"]+)', meta, re.I)
                        if test_charset: return test_charset.group(1)
    if chardet:
        test = chardet.detect(data)
        if 'encoding' in test: return test['encoding']
    return sys.getdefaultencoding()

## URL RELATED

def do_url(type, url, myheaders={}, postdata={}, keyfile=None, certfile="", port=80):
    headers = {'Content-Type': 'application/x-www-form-urlencoded', 'Accept': 'text/plain; text/html', 'User-Agent': useragent()}
    headers.update(myheaders)
    urlparts = urllib.parse.urlparse(url)
    if keyfile: connection = http.client.HTTPSConnection(urlparts[1], port, keyfile, certfile)
    else: connection = http.client.HTTPConnection(urlparts[1])
    postdata = urllib.parse.urlencode(postdata)
    logging.warn('%s %s' % (type, url))
    connection.request(type, urlparts[2], postdata, headers)
    resp = connection.getresponse()
    return resp

## TO/FROM

def to_enc(what, encoding='utf-8'):
    if not what: what=  ""
    w = str(what)
    return w.encode(encoding)

def from_enc(txt, encoding='utf-8', what=""):
    if not txt: txt = ""
    if type(txt) == str: return txt
    try: return txt.decode(encoding)
    except UnicodeDecodeError: return decodeperchar(txt, encoding, what)

## PER CHARACTER

def decode_char(txt, encoding='utf-8', what=""):
    res = [] ; nogo = []
    for i in txt:
        try: res.append(i.decode(encoding))
        except UnicodeDecodeError:
            if i not in nogo: nogo.append(i)
    if nogo: logging.warn("nogo: %s" % " ".join(nogo))
    return "".join(res)

## OPTIONS

def make_opts():
    parser = optparse.OptionParser(usage='usage: %prog [options]', version=__version__)
    for option in options:
        type, default, dest, help = option[2:]
        if "store" in type:
            try: parser.add_option(option[0], option[1], action=type, default=default, dest=dest, help=help)
            except Exception as ex: logging.error("error: %s - option: %s" % (str(ex), option)) ; continue
        else:
            try: parser.add_option(option[0], option[1], type=type, default=default, dest=dest, help=help)
            except Exception as ex: logging.error("error: %s - option: %s" % (str(ex), option)) ; continue
    return parser.parse_args()

